#-----------------------------------------------------------------------
# This python script updates the CEA system app (commsvc.ear)
# from directory <washome>/feature_packs/cea/systemApps/
# to all existing deployments.
#
# Args
# - Fully-qualified path to commsvc.ear
#
# For example:
#  /opt/WAS70/bin/wsadmin.sh
#    -lang jython
#    -conntype NONE
#    -f /opt/WAS70/properties/version/nif/update.cea/config/install/actions/updateCOMMSVC.py
#    /opt/WAS70/feature_packs/cea/systemApps/commsvc.ear
#
#-----------------------------------------------------------------------
# Sample wsadmin commands implemented herein
#
# 1.  Start wsadmin                       wsadmin -lang jython
# 2.  Save the deployment targets wsadmin>targets = AdminApp.listModules('commsvc',['-server'])
#                                         wsadmin>targets = targets.split('#')
# 3.  Update the application              wsadmin>AdminApp.install('<washome>/feature_packs/cea/systemApps/commsvc.ear',['-update','-appname', 'commsvc', '-skipPreparation', '-zeroEarCopy',
#                                                    '-installed.ear.destination', '<washome>/feature_packs/cea/systemApps/'])
#                                         Note: Change both instances of <washome> to the actual path.
# 4.  Update the deployment targets       wsadmin>AdminApp.edit('commsvc', '[ -MapModulesToServers [[ "commsvc" commsvc.rest.war,WEB-INF/sip.xml ' + targets[2] + ']]]' )
# 5.  Save the configuration              wsadmin>AdminConfig.save()
#
#-----------------------------------------------------------------------
import os.path
import sys
import time

# Define False, True
(False,True)=(0,1)

def _splitlines(s):
  rv = [s]
  if '\r' in s:
    rv = s.split('\r\n')
  elif '\n' in s:
    rv = s.split('\n')
  if rv[-1] == '':
    rv = rv[:-1]
  return rv

def getSopTimestamp():
    """Returns the current system timestamp in a nice format."""
    # Assemble the formatting string in pieces, so that CMVC does not interpret
    # the strings as special keywords and substitute them upon extraction.
    formatting_string = "[" + "%" + "Y-" + "%" + "m" + "%" + "d-" + "%" + "H" + "%" + "M-" + "%" + "S00]"
    return time.strftime(formatting_string)

def sop(methodname,message):
    """Prints the specified method name and message with a nicely formatted timestamp.
    (sop is an acronym for System.out.println() in java)"""
    timestamp = getSopTimestamp()
    print "%s %s %s" % (timestamp, methodname, message)

def stringListToList(stringList):
    """Helper method converts a string which looks like a list into a list of strings.

    Expects string input of the form [(cells/ding/...0669) (cells/ding/...6350) (cells/ding/...6291)]
    Returns string output of the form  ['(cells/ding/...0669)', '(cells/ding/...6350)', '(cells/ding/...6291)']
    Note: string.split does not work because of the square brackets."""
    m = "stringListToList:"
    sop(m,"Entry. stringList=%s" % ( stringList ))

    # Input check.
    if not (stringList.startswith( '[' ) and stringList.endswith( ']' )):
        raise m + " ERROR: stringList does not start and end with square brackets. stringList=%s" % ( m, stringList, )

    # Strip off the leading and trailing square brackets.
    stringList = stringList[1:(len(stringList) - 1)]

    # Get rid of the leading and trailing square brackets.
    listList = stringList.split()

    sop(m,"Exit. listList=%s" % (listList))
    return listList

def getObjectAttribute(objectid, attributename):
    """Return the value of the named attribute of the config object with the given ID.
    If there's no such attribute, returns None.
    If the attribute value looks like a list, converts it to a real python list.
    """
    #sop("getObjectAttribute:","AdminConfig.showAttribute(%s, %s)" % ( repr(objectid), repr(attributename) ))
    result = AdminConfig.showAttribute(objectid, attributename)
    if result != None and result.startswith("[") and result.endswith("]"):
        # List looks like "[value1 value2 value3]"
        result = _splitlist(result)
    return result

def getNumberFromId(obj_id):
    """Returns the long decimal number (as a string) from the end of a wsadmin object id string.

    For example, returns 1157676511879 from the following id:
    PAP_1(cells/ding6Cell01|coregroupbridge.xml#PeerAccessPoint_1157676511879)

    Returns the original id string if the ID string can not be parsed.
    """
    longnum = obj_id
    if longnum.endswith(')'):
        # Strip off the trailing parenthesis
        len_orig = len(longnum)
        longnum = longnum[:len_orig-1]
        # Find the index of the last underscore.
        ix = longnum.rfind('_')
        if -1 != ix:
            # Extract the number.
            longnum = longnum[ix+1:]
    return longnum

def getObjectIdSansNum(obj_id):
    """Returns the object_id without the long decimal number at the end.
    This can be useful to correlate config objects which have been deleted and restored,
    which causes the number to be changed.
    For example, given this obj_id,
        PAP_1(cells/ding6Cell01|coregroupbridge.xml#PeerAccessPoint_1157676511879)
    returns
        PAP_1(cells/ding6Cell01|coregroupbridge.xml#PeerAccessPoint)
    """
    longnum = ( "_%s" % ( getNumberFromId(obj_id) ))
    new_id = obj_id.replace(longnum,"")
    return new_id

def getAppValueFromView(appname,taskname,keystring):
    """Returns the current value for the specified app, task, and key.
    For example, getAppValueFromView(appname,'-MapWebModToVH','Virtual host:')"""
    m = "getAppValueFromView:"
    sop(m,"Entry. appname=%s taskname=%s keystring=%s" % (appname,taskname,keystring))

    # Get the current setting
    try:
        verboseView = AdminApp.view(appname, taskname)
    except:
        raise m + "ERROR. Could not query the current setting. appname=%s taskname=%s" % ( appname,taskname )
    sop(m,"verboseView=%s" % (verboseView))

    # Convert to a list of strings
    viewList = _splitlines(verboseView)
    #sop(m,"viewList=%s" % ( repr(viewList) ))

    # Find the element with the desired string.
    desiredValue = "NOT_FOUND"
    for view in viewList:
        #sop(m,"Considering view=%s" % ( view ))

        # Find the key string.
        ixKey = view.find(keystring)
        if -1 == ixKey:
            #sop(m,"No match.")
            continue
        #sop(m,"Found match. ixKey=%i len(view)=%i" % ( ixKey, len(view) ))

        # Strip off the key string in front.
        desiredValue = view[(ixKey + len(keystring)):].strip()
        #sop(m,"Stripped prefix. desiredValue=>>>%s<<<" % ( desiredValue ))

    if "NOT_FOUND" == desiredValue:
        raise m + "ERROR. Could not find the current setting. appname=%s" % ( appname )

    sop(m,"Exit. Returning desiredValue=%s" % ( desiredValue, ))
    return desiredValue

def getAppMapWebModToVH(appname):
    """Returns the current value of MapWebModToVH for the specified app."""
    return getAppValueFromView(appname,'-MapWebModToVH','Virtual host:')

def getAppCtxRootForWebMod(appname):
    """Returns the current value of CtxRootForWebMod for the specified app."""
    return getAppValueFromView(appname,'-CtxRootForWebMod','Context Root:')

def deploymentExists(deploymentname):
    """Returns True(1) or False(0) depending on whether any deployments exist for the specified app."""
    m = "deploymentExists:"
    try:
        deployments = AdminConfig.getid("/Deployment:%s/" % ( deploymentname ))
        # Note: deployments may be a two character string, quote quote:  "''".
        if None != deployments and 2 < len(repr(deployments)):
            sop(m,"Returning True(1). deploymentname=%s deployments=%s." % ( deploymentname, repr(deployments) ))
            return True
        else:
            sop(m,"Returning False(0). deploymentname=%s deployments=%s." % ( deploymentname, repr(deployments) ))
            return False
    except:
        sop(m,"Returning False(0). Could not get id for deployment=%s." % ( deploymentname ))
        return False


def getDeploymentInfoList(deploymentname):
    """Returns a python list of dictionaries, where each dictionary describes a deployment.
    For example, returns this:
        [  { 'name': 'commsvc',
             'target': 'server2(cells/ndcell/applications/commsvc.ear/deployments/commsvc|deployment.xml#ServerTarget_1265821525792)',
             'target_mapping': '(cells/ndcell/applications/commsvc.ear/deployments/commsvc|deployment.xml#DeploymentTargetMapping_1265821525792)',
             'enable': 'true' },
        ]
    Returns an empty list if no deployments are found.
    """
    m = "getDeploymentInfoList:"
    sop(m,"Entry. deploymentname=%s" % ( deploymentname ))
    infoList = []

    rc = 0
    found = 0
    deployments = AdminConfig.getid("/Deployment:%s/" % ( deploymentname ))
    if (len(deployments) > 0) :
        deploymentObj = AdminConfig.showAttribute(deployments, 'deployedObject')
        sop(m,"deploymentObj=%s" % ( repr(deploymentObj) ))

        # First get the Target Mappings.  These are the objects where we set enabled/disabled.
        rawTargetMappings = AdminConfig.showAttribute(deploymentObj, 'targetMappings')
        # Convert the single string to a real python list containing strings.
        targetMappingList = stringListToList(rawTargetMappings)
        sop(m, "targetMappingList=%s" % ( repr(targetMappingList) ))

        # Handle each target mapping...
        for targetMapping in targetMappingList:
            attr_target = getObjectAttribute(targetMapping,"target")
            attr_enable = getObjectAttribute(targetMapping,"enable")
            sop(m,"targetMapping=%s attr_target=%s attr_enable=%s" % ( targetMapping, attr_target, attr_enable ))

            infoDict = { 'name': deploymentname,
                         'target': attr_target,
                         'target_mapping': targetMapping,
                         'enable': attr_enable,
                       }
            infoList.append(infoDict)
    else:
        sop(m, "No deployments found.")

    sop(m,"Exit. Returning infoList=%s" % ( repr(infoList) ))
    return infoList

def setDeploymentAutoStartList( infoList ):
    """Sets an application to start automatically, when the server starts.
    Specify a python list of dictionaries, where each dictionary describes a deployment
    and each dictionary contains the target mapping and desired enabled state.
    For example, specify this:
        [  { 'name': 'commsvc',
             'target': 'server2(cells/ndcell/applications/commsvc.ear/deployments/commsvc|deployment.xml#ServerTarget_1265821525792)',
             'target_mapping': '(cells/ndcell/applications/commsvc.ear/deployments/commsvc|deployment.xml#DeploymentTargetMapping_1265821525792)',
             'enable': 'true' },
        ]
    Returns the number of deployments which were found and set successfully.
    """
    m = "setDeploymentAutoStartList:"
    sop(m,"Entry. infoList=%s" % ( repr(infoList) ))
    numSet = 0

    for deployment in infoList:
        sop(m,"Handling deployment=%s" % ( repr(deployment) ))

        # Check for errors

        if 'target_mapping' not in deployment.keys():
            sop(m,"Warning: Invocation Error. target_mapping is not specified. deployment=%s" % ( repr(deployment) ))
            continue
        target_mapping = deployment['target_mapping']

        if 'enable' not in deployment.keys():
            sop(m,"Warning: Invocation Error. enable is not specified. deployment=%s" % ( repr(deployment) ))
            continue
        enable = deployment['enable']

        if 'true' != enable and 'false' != enable:
            sop(m,"Warning: Invocation Error. enable is not 'true' or 'false'. deployment=%s" % ( repr(deployment) ))
            continue

        # Set the value

        valueString = '[[enable "%s"]]' % ( enable )
        sop(m,"Setting autostart on desired deployment target. target_mapping=%s and value=%s" % ( target_mapping, valueString ))
        AdminConfig.modify(target_mapping, valueString)
        numSet += 1

    sop(m,"Exit. Set %i deployments." % ( numSet ))
    return numSet

def correlateDeploymentAutoStart( oldList, newList ):
    """Compares two lists of information about deployment autostart,
    and propagates autostart information from the old list to the new.
    This method is useful to restore autostart settings when an app is updated
    and the ID numbers change.

    Specify two python lists of dictionaries, where each dictionary describes a deployment
    and each dictionary contains the target mapping and desired enabled state.
    For example, specify this:
        [  { 'name': 'commsvc',
             'target': 'server2(cells/ndcell/applications/commsvc.ear/deployments/commsvc|deployment.xml#ServerTarget_1265821525792)',
             'target_mapping': '(cells/ndcell/applications/commsvc.ear/deployments/commsvc|deployment.xml#DeploymentTargetMapping_1265821525792)',
             'enable': 'true' },
        ]
    Returns the number of deployments correlated and changed.
    """
    m = "correlateDeploymentAutoStart:"
    sop(m,"Entry. old=%s new=%s" % ( repr(oldList), repr(newList) ))
    numSet = 0

    # Process all elements in the new list.
    for newDep in newList:
        sop(m,"Handling newDep=%s" % ( repr(newDep) ))
        # Find the corresponding element in the old list.
        for oldDep in oldList:
            sop(m,"Comparing to oldDep=%s" % ( repr(oldDep) ))
            if newDep['name'] == oldDep['name'] and \
               getObjectIdSansNum(newDep['target']) == getObjectIdSansNum(oldDep['target']) and \
               getObjectIdSansNum(newDep['target_mapping']) == getObjectIdSansNum(oldDep['target_mapping']):
                sop(m,"Found match. Setting enable %s" % ( oldDep['enable'] ))
                newDep['enable'] = oldDep['enable']
                numSet += 1
                break
            else:
                sop(m,"No match")

    sop(m,"Exit. Correlated %i deployments." % ( numSet ))
    return numSet


def updateCOMMSVCApp(ear, saveOnSuccess):
    """Updates the CEA system app (commsvc.ear)
       to all existing deployments.
       Args:
           ear - fully-qualified path to commsvc.ear directory
           saveOnSuccess - True/False
       """
    m = "updateCOMMSVCApp:"
    sop(m,"Entry. ear=%s saveOnSuccess=%s" % ( ear,repr(saveOnSuccess) ))

    # Exit gracefully if the commsvc app is not installed
    if not deploymentExists('commsvc'):
        sop(m,"Exit. Deployment commsvc does not exist. No action required.")
        return

    # Confirm the specified expanded-ear directory exists.
    if not os.path.exists(ear):
        sop(m,"Exit. %s does not exist." % ( ear ))
        return
    if os.path.isfile(ear):
        sop(m,"Exit. %s is not a directory." % ( ear ))
        return

    # Query old values - before updating the app
    oldDeploymentInfoList = getDeploymentInfoList('commsvc')
    oldContextRoot = getAppCtxRootForWebMod('commsvc')
    oldVirtualHost = getAppMapWebModToVH('commsvc')

    # Update the app.
    try:
        rawTargets = AdminApp.listModules('commsvc',['-server'])
        sop(m,"rawTargets=%s" % (rawTargets))
    except:
        sop(m,"Exit. COMMSVC is not installed on any targets.")
        return

    # Convert the #-delimited-string into a python list
    targetList = rawTargets.split('#')

    # Get the list of deployments from the 3rd element
    deploymentTargets = targetList[2]
    sop(m,"deploymentTargets=%s" % ( deploymentTargets ))

    # Define arguments for the update command using a python list of strings.
    arglist = ['-update',
               '-appname', 'commsvc',
               '-skipPreparation',
               '-zeroEarCopy',
               '-installed.ear.destination', ear[:-11]]

    # Do the update
    sop(m,"Calling AdminApp.install with arglist=%s" % ( arglist ))
    try:
        AdminApp.install(ear, arglist)
    except:
        sop(m,"Exit. Caught exception calling AdminApp.install: %s" % (sys.exc_info()[0]))
        return

    # Define arguments for the edit command using one long string.
    argstring = '[ -MapModulesToServers [[ "commsvc" commsvc.rest.war,WEB-INF/sip.xml ' + deploymentTargets + ']]]'

    # Do the edit.
    sop(m,"Calling AdminApp.edit with argstring=%s" % ( repr(argstring) ))
    AdminApp.edit('commsvc', argstring)

    # Query info about each deployment after updating the app.
    newDeploymentInfoList = getDeploymentInfoList('commsvc')

    # Restore old values from before the app was updated.
    correlateDeploymentAutoStart(oldDeploymentInfoList,newDeploymentInfoList)
    setDeploymentAutoStartList(newDeploymentInfoList)
    AdminApp.edit('commsvc', '[ -MapWebModToVH [[ commsvc commsvc.rest.war,WEB-INF/web.xml %s ]]]' % (oldVirtualHost) )
    AdminApp.edit('commsvc', '[ -CtxRootForWebMod [[ commsvc commsvc.rest.war,WEB-INF/web.xml %s ]]]' % ( oldContextRoot ))

    # Save
    if saveOnSuccess:
        sop(m,"Saving")
        AdminConfig.save()

    sop(m,"Exit. Success.")


#--------------------------------------------------------------------
# main - The script starts here.
#--------------------------------------------------------------------
m = "updateCOMMSVC.py:"
sop(m,"Entry. FTMID=10F2")

# Get the fully-qualified path to commsvc.ear
# For example, /opt/WAS70/feature_packs/cea/systemApps/commsvc.ear
ear = sys.argv[0]
sop(m,"ear=%s" % ( ear ))

# Update the app on all deployments.
saveOnSuccess = True
updateCOMMSVCApp(ear, saveOnSuccess)

sop(m,"Exit")
